#!/usr/bin/env python
from trackcluster.utils import myexe, is_bin_in, get_file_prefix,del_files,get_file_location
from trackcluster.tracklist import read_bigg, write_bigg, list_to_dic
from trackcluster.pre import tracklist_add_gene,get_gendic_bedinter,group_bigg_by_gene
import os
import argparse

def wrapper_bedtools_intersect2_select(bedfile1,bedfile2,outfile,fraction_bed1, fraction_bed2):
    """
    Using two bedfile to get the intsersection of pairs
    use simular fraction as substract instead of 0.2, the bigg can contain large intron in first intersect

    # -f 0.01= 1% from bigg, -F 0.05, means 5% from gene annotation

    :param bigg_one: the read
    :param bigg_two: the ref
    :param: fraction, may need to inlcude some read track with very long
    :return:
    """
    # generate the bedfile, -r is reciprocal overlap, -s is strandedness/keep the same strand

    # sometime the name check for an intact bed and an non-intact one can result in a blank file
    # add -nonamecheck
    cmd="bedtools intersect -nonamecheck -wa -wb  -s -f {f1} -F {f2} -a {bedfile1} -b {bedfile2}>{out}".format(
        bedfile1=bedfile1, bedfile2=bedfile2, out=outfile, f1=fraction_bed1, f2=fraction_bed2)

    _=myexe(cmd)

    return outfile

def flow_add_gene(wkdir, prefix, bigg_gff_file, bigg_nano_file):
    os.chdir(wkdir)
    # make sure one read one track
    bigg_raw=read_bigg(bigg_nano_file)
    bigg_dedup=list(list_to_dic(bigg_raw).values())
    print("raw bigg number: {}; after dedup:{}".format(len(bigg_raw), len(bigg_dedup)))
    outbed=prefix+"_dedup.bed"
    write_bigg(bigg_dedup, outbed)

    ### get two parts, the gene part and the novel part,

    # the gene part
    out_single = prefix + "_single_inter.bed"
    # write the outfile to disk
    wrapper_bedtools_intersect2_select(outbed, bigg_gff_file, outfile=out_single,
                                       fraction_bed1=0.01, fraction_bed2=0.05)
    out_fusion = prefix + "_fusion_inter.bed"
    # write the outfile to disk
    wrapper_bedtools_intersect2_select(outbed, bigg_gff_file, outfile=out_fusion,
                                       fraction_bed1=0.33, fraction_bed2=0.33)

    read_gene_single = get_gendic_bedinter(out_single)
    read_gene_fusion = get_gendic_bedinter(out_single)
    print("read number in genes:", len(read_gene_single))

    bigg_single = tracklist_add_gene(bigg_dedup, read_gene_single)
    bigg_fusion = tracklist_add_gene(bigg_dedup, read_gene_fusion)

    # cleanup
    del_files([outbed, out_single, out_fusion])

    return bigg_single, bigg_fusion

def flow_fusion_annotation(wkdir,isoform_bed, prefix=None):
    if prefix is None:
        prefix=get_file_prefix(isoform_bed, sep=".")

    os.chdir(wkdir)
    bigg_isoform=read_bigg(isoform_bed)

    # make the fusion class separately
    fusion_d = {}
    for bigg in bigg_isoform:
        genename_l=bigg.geneName.split("||")
        if len(genename_l)>1:
            fusion_d[bigg.name]=genename_l

    # IO for fusion
    with open(prefix+"_fusion.txt", "w") as fw:
        for k, v in fusion_d.items():
            v_str = ";".join(v)
            fw.write(k + "\t" + v_str + "\n")

    return prefix+"_fusion.txt"

def flow_single_annotation(wkdir,isoform_bed, prefix=None):
    if prefix is None:
        prefix=get_file_prefix(isoform_bed, sep=".")

    os.chdir(wkdir)
    bigg_isoform=read_bigg(isoform_bed)

    # make the fusion class separately
    with open(prefix + "_single.txt", "w") as fw:
        for bigg in bigg_isoform:
            if bigg.geneName != 'none':
                genename_l=bigg.geneName.split("||")
                if len(genename_l) == 1:
                    fw.write(bigg.name + "\t" + genename_l[0] + "\n")

    return prefix + "_single.txt"

def concatenate_and_remove(file1, file2, output_file):
    # 打开输出文件，以写入模式
    with open(output_file, 'w') as outfile:
        # 读取并写入第一个文件的内容
        with open(file1, 'r') as f1:
            outfile.write(f1.read())
        # 读取并写入第二个文件的内容
        with open(file2, 'r') as f2:
            outfile.write(f2.read())

if __name__ == '__main__':
# def addgene(self):
    parser = argparse.ArgumentParser(
        description="Used to add gene annotation for read bigg tracks, useful in some analysis, "
                    "the process is included in cluster runs. the new bigg file will be prefix_gene.bed"
    )
    parser.add_argument("-d", "--folder", default=os.getcwd(),
                        help="the folder contains all the seperated tracks in different locus/genes, default is the current dir")
    parser.add_argument("-s", "--sample",
                        help="the bigg format of the read track, with the key of GeneName")
    parser.add_argument("-r", "--reference",
                        help="the bigg format of the reference annotation track")

    args = parser.parse_args()
    args.prefix = get_file_prefix(args.sample, sep=".")

    os.chdir(args.folder)
    bigg_single, bigg_fusion = flow_add_gene(wkdir=args.folder,
                                             prefix=args.prefix,
                                             bigg_gff_file=args.reference,
                                             bigg_nano_file=args.sample,
                                             )
    out_single = args.prefix + "_single_gene.bed"
    out_fusion = args.prefix + "_fusion_gene.bed"
    write_bigg(bigg_single, out_single)
    write_bigg(bigg_fusion, out_fusion)

    fusion_f = flow_fusion_annotation(wkdir=args.folder,
                                      isoform_bed=out_fusion,
                                      prefix=args.prefix)

    single_f = flow_single_annotation(wkdir=args.folder,
                                      isoform_bed=out_single,
                                      prefix=args.prefix)
    output_file = args.prefix + "_gene.bed"
    concatenate_and_remove(single_f, fusion_f, output_file)
    del_files([out_single, out_fusion, single_f, fusion_f])



#!/usr/bin/env python
import re
import argparse
import pysam
import pandas as pd
# from ssw import AlignmentMgr
from pyssw.ssw_wrap import Aligner
from Bio.Seq import Seq
import random
import multiprocessing
from tqdm import tqdm
import time
from SLRanger.visualization import visualize_html

def fasta_to_dict(fasta_path):
    fasta_dict = {}
    with open(fasta_path, 'r') as f:
        current_key = None
        sequence_parts = []
        for line in f:
            line = line.strip()
            if line.startswith('>'):
                if current_key:
                    # 保存上一个序列
                    fasta_dict[current_key] = ''.join(sequence_parts)
                current_key = line[1:]  # 去掉 '>'
                sequence_parts = []
            else:
                sequence_parts.append(line)
        # 保存最后一个序列
        if current_key:
            fasta_dict[current_key] = ''.join(sequence_parts)
    return fasta_dict


def generate_mismatches(kmer, bases=['A', 'T', 'C', 'G']):
    mismatches = set()
    for i in range(len(kmer)):
        for base in bases:
            if base != kmer[i]:
                mismatched_kmer = kmer[:i] + base + kmer[i + 1:]
                mismatches.add(mismatched_kmer)
    return mismatches

def build_mismatch_index(sequences, k):
    mismatch_to_kmer = {}

    for seq_id, seq in sequences.items():
        for i in range(len(seq) - k + 1):
            kmer = seq[i:i + k]
            mismatches = generate_mismatches(kmer)
            for mismatch in mismatches:
                if mismatch not in mismatch_to_kmer:
                    mismatch_to_kmer[mismatch] = kmer

    return mismatch_to_kmer


def extract_kmers(sequences, k):
    kmer_id = {}
    # 遍历每个序列并为每个5-mer分配ID
    for key, seq in sequences.items():
        encoded_seq = []
        for i in range(len(seq) - k + 1):
            kmer = seq[i:i+k]
            encoded_seq.append(kmer)
        kmer_id[key] = encoded_seq
    return kmer_id

def find_matches(query, mismatch_to_kmer, kmer_to_refs, k):
    query_kmers = [query[i:i + k] for i in range(len(query) - k + 1)]
    matches = []

    for query_kmer in query_kmers:
        if query_kmer in mismatch_to_kmer:
            original_kmer = mismatch_to_kmer[query_kmer]
            if original_kmer in kmer_to_refs:
                match_info = kmer_to_refs[original_kmer]
                matches.extend(match_info)

    return matches

def longest_consecutive(nums):
    num_set = set(nums)
    longest_streak = 0
    max_end_value = None

    for num in num_set:
        if num - 1 not in num_set:
            current_num = num
            current_streak = 1

            while current_num + 1 in num_set:
                current_num += 1
                current_streak += 1

            if current_streak > longest_streak:
                longest_streak = current_streak
                max_end_value = current_num

    return longest_streak, max_end_value

def find_best_match(query, mismatch_to_kmer, kmer_to_refs, k):
    encoded_query = [query[i:i+k] for i in range(len(query) - k + 1)]
    ref_positions = []
    max_consecutive = 0
    best_consecutive_end = 0

    for query_kmer in encoded_query:
        if query_kmer in kmer_to_refs:
            if kmer_to_refs.count(query_kmer) > 1:
                ref_positions.extend(index for index, kmer in enumerate(kmer_to_refs) if kmer == query_kmer)
            else:
                ref_positions.append(kmer_to_refs.index(query_kmer))
        elif query_kmer in mismatch_to_kmer:
            original_kmer = mismatch_to_kmer[query_kmer]
            if original_kmer in kmer_to_refs:
                if kmer_to_refs.count(query_kmer) > 1:
                    ref_positions.extend(index for index, kmer in enumerate(kmer_to_refs) if kmer == original_kmer)
                else:
                    ref_positions.append(kmer_to_refs.index(original_kmer))
    max_intersection = len(set(ref_positions))
    if max_intersection > 0:
        max_consecutive, best_consecutive_end = longest_consecutive(ref_positions)

    if max_consecutive > len(query) - k:
    # if max_intersection > max_consecutive:
        max_intersection = max_consecutive

    return max_intersection, max_consecutive, best_consecutive_end


def soft_processed(sequence, strand, cigar):
    """
    根据 strand 和 cigar 信息提取指定长度的序列部分。

    参数:
    - sequence (str): 原始的序列字符串。这个sequence已经被reverse处理过了，与正链一致，所以取序列的时候都从开头取起
    - strand (str): 链方向，"+" 或 "-"。
    - cigar (list of tuples): CIGAR 信息，表示为列表的元组 (operation, length)。
                              例如 [(4, 5), (0, 50)] 表示 5 bp 的 soft clipping，接着是 50 bp 的匹配。
    返回:
    - str: 根据条件截取后的序列。再多加两个碱基在后面
    """
    if not cigar:
        return ""  # Return empty string if no CIGAR info

    if strand == "+":
        if cigar[0][0] == 4:
            soft_clip_length = cigar[0][1]
            # 返回从序列开头截取 soft_clip_length 长度的序列部分
            return sequence[:soft_clip_length + 2]  #
    elif strand == "-":
        if cigar[-1][0] == 4:
            soft_clip_length = cigar[-1][1]
            # 返回从序列开头截取 soft_clip_length 长度的序列部分
            return sequence[:soft_clip_length + 2]  #
    else:
        # 不符合条件时，返回原始序列
        return ""


def ssw_wrapper(seq1, seq2, match=1, mismatch=1, gap_open=1, gap_extend=1):
    """
    parameter are write inside the function
    seq1 is ref and seq2 is query
    # todo : leave a api to change matrix
    """

    ref_seq = str(seq1)
    read_seq = str(seq2)
    # reduce the gap open score from 3 to 1 for nanopore reads
    aligner = Aligner(ref_seq,
                      match, mismatch, gap_open, gap_extend,
                      report_cigar=True, )

    aln = aligner.align(read_seq)  # min_score=20, min_len=10)

    return aln

def consensus(ref, query, cigar_string, ref_shift=0, query_shift = 0):
    """
    to get a concensus sequence as the longest cons of both sequence,
    and fill all disagreement with N
    for example,
    query:AAATA-TAGAA
    ref:  AAACACTA-AA
    cons: AAANANCANAA

    use query to get the con
    then get two chain file, 0 based
    """

    # pattern = re.compile('([0-9]*)([MIDNSHP=X])')
    pattern = re.compile(r"([0-9]+)b?'?([MIDNSHP=X])")
    query_out = []
    ref_out = []
    cons = []

    query_pos = query_shift
    ref_pos = ref_shift

    cigar_pattern = pattern.findall(cigar_string)
    for i, (length, code) in enumerate(cigar_pattern):
        if i == 0 or i == len(cigar_pattern) - 1:
            if code == "S":
                continue

        length = int(length)

        if code == "M":
            for i in range(length):
                q_s = query[query_pos].upper()
                try:
                    r_s = ref[ref_pos].upper()
                except IndexError as e:
                    print
                    "indexerror", ref_pos

                query_out.append(q_s)
                ref_out.append(r_s)

                cons.append(q_s) if q_s == r_s else cons.append("m")

                ref_pos += 1
                query_pos += 1

        elif code in "D":
            for i in range(length):
                r_s = ref[ref_pos]

                ref_out.append(r_s)
                query_out.append("-")
                cons.append("d")

                ref_pos += 1

        elif code in "IHS":
            for i in range(length):
                q_s = query[query_pos]

                query_out.append(q_s)
                ref_out.append("-")
                cons.append("i")
                query_pos += 1

    return "".join(cons)   # "".join(cons), "".join(query_out), "".join(ref_out)

def score_calculate(sw_score, max_intersection, max_consecutive, ref_end, ref_length, seq_start, seq_length, soft_length):
    # factor = get_value(best_end, ref_length - k)

    score_seq = (1 - seq_start/soft_length) # * (seq_length/soft_length)  # 在soft clipping中的位置，越靠头部越好
    # score_ref = 1 - (ref_length - ref_end)/ref_length
    score_ref = 1 - (ref_length - ref_end)/25   #(SL1_len - k)*2
    score_kmer = 0.5 * (max_intersection + max_consecutive - 1)
    score = score_seq * score_ref * (sw_score + score_kmer)

    final_score = max(score, 0)
    return final_score

def length_index(SL, ref_seq, kmer, mismatch_to_kmer, k):
    dict = {}

    for i in range(k, len(ref_seq) + 1):
        ref_seq_s = ref_seq[-i:]
        sw_aln = ssw_wrapper(ref_seq, ref_seq_s)  # score>10,
        sw_score = sw_aln.score
        # ref_start = sw_aln.ref_begin
        # ref_end = sw_aln.ref_end
        read_start = sw_aln.query_begin
        read_end = sw_aln.query_end + 1
        # sw_cigar = sw_aln.cigar_string

        corrected_sequence_sw = ref_seq_s
        # last_three_chars = corrected_sequence_sw[-3:]
        seq_s_length = len(corrected_sequence_sw)
        ref_length = len(ref_seq)
        seq_length = read_end - read_start
        kmer_to_refs = kmer[SL]
        max_intersection, max_consecutive, best_end = find_best_match(corrected_sequence_sw, mismatch_to_kmer,
                                                                      kmer_to_refs, k)
        final_score = score_calculate(sw_score, max_intersection, max_consecutive, ref_length, ref_length,
                                      read_start, seq_s_length, len(ref_seq_s))
        dict[i] = final_score

    return dict

def final_score_process(final_score, ref_length, seq_s_length, length_score):
    if seq_s_length <= ref_length:
        SL_score = final_score * (final_score / length_score[seq_s_length])
        # final_score_normalized = 100 * (final_score / length_score[seq_s_length])
    else:
        SL_score = final_score * (final_score / (seq_s_length / ref_length * length_score[ref_length]))
        # final_score_normalized = 100 * (final_score / (seq_s_length / ref_length * length_score[ref_length]))
    return SL_score # final_score_normalized

def random_score(random_sequences_dict, random_kmer, random_mismatch_to_kmer, length_score, corrected_sequence, k):
    random_sw_score_max = 0
    random_final_score_max = 0
    random_SL_score_max = 0
    for key, random_seq in random_sequences_dict.items():
        random_length_score = length_score
        random_aln_sw = ssw_wrapper(random_seq, corrected_sequence)
        random_sw_score = random_aln_sw.score
        random_read_start = random_aln_sw.query_begin
        random_read_end = random_aln_sw.query_end + 1
        corrected_sequence_sw = corrected_sequence[random_read_start:random_read_end]
        seq_s_length = len(corrected_sequence_sw)

        if seq_s_length >= 5:
            random_ref_end = random_aln_sw.ref_end + 1
            random_ref_length = len(random_seq)
            kmer_to_refs = random_kmer[key]
            max_intersection, max_consecutive, best_end = find_best_match(corrected_sequence_sw,
                                                                          random_mismatch_to_kmer,
                                                                          kmer_to_refs, k)
            random_final_score = score_calculate(random_sw_score, max_intersection, max_consecutive,
                                                 random_ref_end, random_ref_length, random_read_start,
                                                 seq_s_length, len(corrected_sequence))
            random_SL_score = final_score_process(random_final_score, random_ref_length,
                                                  seq_s_length, random_length_score)
            if random_SL_score > random_SL_score_max:
                random_sw_score_max = random_sw_score
                random_final_score_max = random_final_score
                random_SL_score_max = random_SL_score

    return random_sw_score_max, random_final_score_max, random_SL_score_max

def update(*a):
    if a[0] is not None:
        outfile.write(a[0])
    pbar.update(1)

def calculation_per_process(item,sl_dict,length_scores,random_sequences_dict ,random_kmer,random_mismatch_to_kmer,k,kmer,mismatch_to_kmer):
    query_name = item[0]
    query_seq = item[1]
    strand = item[2]

    corrected_sequence = soft_processed(query_seq, strand, item[3])
    aligned_len = item[4]

    if corrected_sequence is not None:
        soft_length = len(corrected_sequence)
        mes = query_name + '\t' + strand + '\t' + str(soft_length) + '\t' + str(aligned_len) + '\t'  # with soft_processed
    else:
        soft_length = 0
        mes = query_name + '\t' + strand + '\t' + 'NA' + '\t' + str(aligned_len) + '\t'  # with soft_processed

    if soft_length < k:  # 太短了就不要了
        SL_sw_dict = {'read_end': 'NA', 'query_length': 'NA',
                      'consensus': 'NA', 'random_sw_score': 'NA',
                      'random_final_score': 'NA', 'random_SL_score': 'NA',
                      'sw_score': 'NA', 'final_score': 'NA',
                      'SL_score': 'NA'}
        mes = mes + '\t'.join([str(value) for value in SL_sw_dict.values()]) + '\t' + 'random' + "\n"
        return mes

    ### use SL1 seq for SW check
    SL_sw_dict = {}
    for SL, SEQ in sl_dict.items():
        length_score = length_scores[SL]
        # corrected_sequence = 'CAAG'
        sw_aln = ssw_wrapper(SEQ, corrected_sequence)  # score>10,
        sw_score = sw_aln.score
        ref_start = sw_aln.ref_begin
        ref_end = sw_aln.ref_end + 1
        read_start = sw_aln.query_begin
        read_end = sw_aln.query_end + 1
        sw_cigar = sw_aln.cigar_string

        corrected_sequence_sw = corrected_sequence[read_start:read_end]

        seq_s_length = len(corrected_sequence_sw)
        random_sw_score_max, random_final_score_max, random_SL_score_max = random_score(random_sequences_dict,
                                                                                        random_kmer,
                                                                                        random_mismatch_to_kmer,
                                                                                        length_scores['SL1'],
                                                                                        corrected_sequence, k)
        if seq_s_length >= k:
            sw_ref = SEQ[ref_start:ref_end]

            # print("\n".join(consensus(corrected_sequence_sw, sw_ref, sw_cigar, 0)))
            # last_three_chars = corrected_sequence_sw[-3:]
            con_seq = consensus(sw_ref, corrected_sequence_sw, sw_cigar, 0)
            ref_length = len(SEQ)
            kmer_to_refs = kmer[SL]
            max_intersection, max_consecutive, best_end = find_best_match(corrected_sequence_sw, mismatch_to_kmer,
                                                                          kmer_to_refs, k)
            final_score = score_calculate(sw_score, max_intersection, max_consecutive, ref_end, ref_length,
                                          read_start, seq_s_length,
                                          soft_length)  # sw_score, max_intersection, max_consecutive, ref_end, ref_length, seq_start, seq_length, soft_length
            SL_score = final_score_process(final_score, ref_length, seq_s_length, length_score)

            SL_sw_dict[SL] = {'read_end': read_end,
                              'query_length': seq_s_length,
                              'consensus': con_seq,
                              'random_sw_score': random_sw_score_max,
                              'random_final_score': round(random_final_score_max, 2),
                              'random_SL_score': round(random_SL_score_max, 2),
                              'sw_score': sw_score,
                              'final_score': round(final_score, 2),
                              'SL_score': round(SL_score, 2)
                              }
        else:
            SL_sw_dict[SL] = {'read_end': read_end,
                              'query_length': None,
                              'consensus': None,
                              'random_sw_score': 0,
                              'random_final_score': 0,
                              'random_SL_score': 0,
                              'sw_score': 0,
                              'final_score': 0,
                              'SL_score': 0
                              }

    SL_sw_df = pd.DataFrame.from_dict(SL_sw_dict, orient='index')
    SL_sw_df_s = SL_sw_df[SL_sw_df['SL_score'] > SL_sw_df['random_SL_score']]
    sl_max = SL_sw_df_s['SL_score'].max()
    # filtered_df = SL_sw_df_s[SL_sw_df_s['SL_score'] == sl_max]
    filtered_df = SL_sw_df_s[SL_sw_df_s['SL_score'] == sl_max]

    if len(filtered_df) > 1:
        mes = mes + '\t'.join([str(value) for value in filtered_df.iloc[0]]) + '\t' + filtered_df.index[
            0] + '_unknown' + "\n"
    elif len(filtered_df) == 1:
        mes = mes + '\t'.join([str(value) for value in filtered_df.iloc[0]]) + '\t' + filtered_df.index[0] + "\n"
    else:
        filtered_df_s = SL_sw_df[SL_sw_df['SL_score'] == SL_sw_df['SL_score'].max()]
        mes = mes + '\t'.join([str(value) for value in filtered_df_s.iloc[0]]) + '\t' + 'random' + "\n"
    return mes


def main(args):
    global outfile, pbar
    """
    NM comparison between SL1 and SL2
    read reads in bam
    write out a dataframe that including:
        query name, 22nt sequence, SW score, SL1 score, SL2 score, SL1 cigar, SL2 cigar ,SL type
        query name, selected 22nt sequence with soft clipping...
    """
    sl_dict = fasta_to_dict(args.ref)

    # 生成10个长度为SL mean长度的碱基的随机序列
    ref_lengths = [len(key) for key in sl_dict.values()]
    random_seq_len = round(sum(ref_lengths) / len(ref_lengths))
    random.seed(826)
    random_sequences_dict = {}
    for i in range(10):
        random_sequence = ''.join([random.choice('AGTC') for _ in range(random_seq_len)])
        random_sequences_dict[i] = random_sequence
    min_len = 7
    k = 5
    kmer = extract_kmers(sl_dict, k)
    mismatch_to_kmer = build_mismatch_index(sl_dict, k)
    random_kmer = extract_kmers(random_sequences_dict, k)
    random_mismatch_to_kmer = build_mismatch_index(random_sequences_dict, k)

    length_scores = {}
    for SL, SEQ in sl_dict.items():
        length_score = length_index(SL, SEQ, kmer, mismatch_to_kmer, k)
        length_scores[SL] = length_score

    bam_file = pysam.AlignmentFile(args.bam, 'rb')
    timestamp = int(time.time())
    tmp_output_name = f"tmp_{timestamp}.csv"
    outfile = open(tmp_output_name, "w")
    outfile.write("query_name\tstrand\tsoft_length\taligned_length\tread_end\tquery_length\tconsensus\trandom_sw_score\trandom_final_score\trandom_SL_score\tsw_score\tfinal_score\tSL_score\tSL_type\n")

    # 迭代每个read
    print('Loading the BAM file')
    bam_list=[]
    for read in bam_file.fetch():
        if read.is_supplementary or read.is_secondary:
            continue
        ### sequence extract
        query_name = read.query_name     # 获取query的名称
        full_query_sequence = read.query_sequence  # 获取query的序列
        if read.is_reverse:
            query_seq = str(Seq(full_query_sequence).reverse_complement())   # 序列映射到负链，需要反向互补处理
            strand = '-'
        else:
            query_seq = full_query_sequence    # 序列映射到正链，直接使用
            strand = '+'
        bam_list.append([query_name,query_seq,strand,read.cigartuples,read.query_alignment_length])
    pbar = tqdm(total=len(bam_list), position=0, leave=True)
    with multiprocessing.Pool(processes=args.cpu) as pool:
        for item in bam_list:
            pool.apply_async(calculation_per_process, args=(item,sl_dict,length_scores,random_sequences_dict,random_kmer,random_mismatch_to_kmer,k,kmer,mismatch_to_kmer), callback=update)
        pool.close()
        pool.join()
    pbar.close()
    outfile.close()
    df = pd.read_csv(tmp_output_name,sep='\t')
    df.sort_values(by=['query_name'], inplace=True)
    df.to_csv(args.output, index=False,sep='\t')
    if args.visualization:
        visualize_html(args.output, args.cutoff)
    print('Finished')

if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description="help to know spliced leader and distinguish SL1 and SL2")
    parser.add_argument("-r", "--ref", type=str, required=True, help="SL reference")
    parser.add_argument("-b", "--bam", type=str, required=True, help="input the bam file")
    parser.add_argument("-o", "--output", type=str, default="SLRanger.txt",
                        help="output file")
    parser.add_argument( "--visualization", action='store_true', help='Turn on the visualization mode')
    parser.add_argument("-t", "--cpu", type=int,
                        default=4, help="CPU number")
    parser.add_argument("-c", "--cutoff", type=float, default=4, help="cutoff of high confident SL sequence")
    args = parser.parse_args()
    main(args)
